import logging

from keras import optimizers
import keras.backend as K
from keras.layers import Dense
from keras.layers import Input
from keras.layers.merge import concatenate
from keras.models import Model
from keras.optimizers import SGD
from keras.regularizers import l2
import numpy as np
from sklearn.utils import check_random_state
import tensorflow as tf

from csrank.constants import allowed_dense_kwargs
from csrank.layers import create_input_lambda
from csrank.layers import DeepSet
from csrank.learner import Learner
from csrank.util import print_dictionary

__all__ = ["FATENetwork", "FATENetworkCore"]


class FATENetworkCore(Learner):
    def __init__(
        self,
        n_hidden_joint_layers=32,
        n_hidden_joint_units=32,
        activation="selu",
        kernel_initializer="lecun_normal",
        kernel_regularizer=l2(0.01),
        optimizer=SGD(lr=1e-4, nesterov=True, momentum=0.9),
        batch_size=256,
        random_state=None,
        **kwargs,
    ):
        """
            Create a FATE-network architecture.
            Training and prediction complexity is linear in the number of objects.

            Parameters
            ----------
            n_hidden_joint_layers : int
                Number of joint layers.
            n_hidden_joint_units : int
                Number of hidden units in each joint layer
            activation : string or function
                Activation function to use in the hidden units
            kernel_initializer : function or string
                Initialization function for the weights of each hidden layer
            kernel_regularizer : function or string
                Regularizer to use in the hidden units
            optimizer : string or function
                Stochastic gradient optimizer
            batch_size : int
                Batch size to use for training
            random_state : int or object
                Numpy random state
            **kwargs
                Keyword arguments for the hidden units
        """
        self.logger = logging.getLogger(FATENetworkCore.__name__)
        self.random_state = random_state

        self.n_hidden_joint_layers = n_hidden_joint_layers
        self.n_hidden_joint_units = n_hidden_joint_units

        self.activation = activation
        self.kernel_initializer = kernel_initializer
        self.kernel_regularizer = kernel_regularizer
        self.batch_size = batch_size
        self.optimizer = optimizers.get(optimizer)
        self._optimizer_config = self.optimizer.get_config()
        self.joint_layers = None
        self.scorer = None
        keys = list(kwargs.keys())
        for key in keys:
            if key not in allowed_dense_kwargs:
                del kwargs[key]
        self.kwargs = kwargs
        self._construct_layers(
            activation=self.activation,
            kernel_initializer=self.kernel_initializer,
            kernel_regularizer=self.kernel_regularizer,
            **self.kwargs,
        )

    def _construct_layers(self, **kwargs):
        """
            Construct basic layers shared by all ranking algorithms:
             * Joint dense hidden layers
             * Output scoring layer

            Connecting the layers is done in join_input_layers and will be done in implementing classes.

            Parameters
            ----------
            **kwargs
                Keyword arguments passed into the joint layers
        """
        self.logger.info(
            "Construct joint layers hidden units {} and layers {} ".format(
                self.n_hidden_joint_units, self.n_hidden_joint_layers
            )
        )
        # Create joint hidden layers:
        self.joint_layers = []
        for i in range(self.n_hidden_joint_layers):
            self.joint_layers.append(
                Dense(
                    self.n_hidden_joint_units, name="joint_layer_{}".format(i), **kwargs
                )
            )

        self.logger.info("Construct output score node")
        self.scorer = Dense(
            1,
            name="output_node",
            activation="linear",
            kernel_regularizer=self.kernel_regularizer,
        )

    def join_input_layers(self, input_layer, *layers, n_layers, n_objects):
        """
            Accepts input tensors and an arbitrary number of feature tensors and concatenates them into a joint layer.
            The input layers need to be given separately, because they need to be iterated over.

            Parameters
            ----------
            input_layer : input tensor (n_objects, n_features)
            layers : tensors
                A number of tensors representing feature representations
            n_layers : int
                Number of hidden set layers
            n_objects : int
                Number of objects
        """
        self.logger.debug("Joining set representation and joint layers")
        scores = []

        inputs = [create_input_lambda(i)(input_layer) for i in range(n_objects)]

        for i in range(n_objects):
            if n_layers >= 1:
                joint = concatenate([inputs[i], *layers])
            else:
                joint = inputs[i]
            for j in range(self.n_hidden_joint_layers):
                joint = self.joint_layers[j](joint)
            scores.append(self.scorer(joint))
        scores = concatenate(scores, name="final_scores")
        self.logger.debug("Done")

        return scores

    def set_tunable_parameters(
        self,
        n_hidden_joint_units=32,
        n_hidden_joint_layers=2,
        reg_strength=1e-4,
        learning_rate=1e-3,
        batch_size=128,
        **point,
    ):
        self.n_hidden_joint_layers = n_hidden_joint_layers
        self.n_hidden_joint_units = n_hidden_joint_units
        self.kernel_regularizer = l2(reg_strength)
        self.batch_size = batch_size
        # Hack to fix memory leak:
        self.optimizer = self.optimizer.from_config(self._optimizer_config)
        K.set_value(self.optimizer.lr, learning_rate)

        self._construct_layers(
            activation=self.activation,
            kernel_initializer=self.kernel_initializer,
            kernel_regularizer=self.kernel_regularizer,
            **self.kwargs,
        )

        if len(point) > 0:
            self.logger.warning(
                "This ranking algorithm does not support"
                " tunable parameters"
                " called: {}".format(print_dictionary(point))
            )


class FATENetwork(FATENetworkCore):
    def __init__(self, n_hidden_set_layers=1, n_hidden_set_units=1, **kwargs):
        """
            Create a FATE-network architecture.
            Training and prediction complexity is linear in the number of objects.

            Parameters
            ----------
            n_hidden_set_layers : int
                Number of hidden set layers.
            n_hidden_set_units : int
                Number of hidden units in each set layer
            **kwargs
                Keyword arguments for the hidden set units
        """
        FATENetworkCore.__init__(self, **kwargs)
        self.logger_gorc = logging.getLogger(FATENetwork.__name__)

        self.n_hidden_set_layers = n_hidden_set_layers
        self.n_hidden_set_units = n_hidden_set_units
        self.model = None
        self.set_layer = None
        self._create_set_layers(
            activation=self.activation,
            kernel_initializer=self.kernel_initializer,
            kernel_regularizer=self.kernel_regularizer,
        )
        self.is_variadic = True
        self.hash_file = None

    def _create_set_layers(self, **kwargs):
        """
            Create layers for learning the representation of the query set. The actual connection of the layers is done
            during fitting, since we do not know the size(s) of the set(s) in advance.
        """
        self.logger_gorc.info(
            "Creating set layers with set units {} set layer {} ".format(
                self.n_hidden_set_units, self.n_hidden_set_layers
            )
        )
        if self.n_hidden_set_layers >= 1:
            self.set_layer = DeepSet(
                units=self.n_hidden_set_units, layers=self.n_hidden_set_layers, **kwargs
            )
        else:
            self.set_layer = None

    @staticmethod
    def _bucket_frequencies(X, min_bucket_size=32):
        """
            Calculates the relative frequency of each ranking bucket.

            Parameters
            ----------
            X : dict
                map from n_objects to object queries
            min_bucket_size : int
                Minimum number of instances for a query size to be considered for
                the frequency calculation

            Returns
            -------
            freq : dict
                map from n_objects to frequency in float

        """
        freq = dict()
        total = 0.0
        for n_objects, arr in X.items():
            n_instances = arr.shape[0]
            if n_instances >= min_bucket_size:
                freq[n_objects] = n_instances
                total += freq[n_objects]
            else:
                freq[n_objects] = 0
        for n_objects in freq.keys():
            freq[n_objects] /= total
        return freq

    def _construct_models(self, buckets):
        models = dict()
        n_features = self.n_object_features_fit_

        for n_objects in buckets.keys():
            model = self.construct_model(n_features, n_objects)
            models[n_objects] = model
        return models

    def get_weights(self, n_objects=None):
        if self.is_variadic:
            if n_objects is not None:
                weights = self.models_[n_objects].get_weights()
            else:
                weights = self.models_[n_objects].get_weights()
        else:
            weights = self.model.get_weights()
        return weights

    def set_weights(self, weights, n_objects=None):
        if self.is_variadic:
            if n_objects is not None:
                self.models_[n_objects].set_weights(weights)
            else:
                self.models_[0].set_weights(weights)
        else:
            self.model.set_weights(weights)

    def _fit(
        self,
        X=None,
        Y=None,
        generator=None,
        epochs=35,
        inner_epochs=1,
        callbacks=None,
        validation_split=0.1,
        verbose=0,
        global_lr=1.0,
        global_momentum=0.9,
        min_bucket_size=500,
        refit=False,
        optimizer=None,
        **kwargs,
    ):
        """
            Fit a generic FATE-network model.

            This is not intended for direct use. Instead, you should use one of
            the domain-specific subclasses such as `FATEChoiceFuntion` or
            `FATEObjectRanker` instead.

            Parameters
            ----------
            X : numpy array or dict
                Feature vectors of the objects
                (n_instances, n_objects, n_features) if numpy array or map from n_objects to numpy arrays
            Y : numpy array or dict
                The exact semantics are domain dependent and should be
                described in the relevant subclasses.
            epochs : int
                Number of epochs to run if training for a fixed query size or
                number of epochs of the meta gradient descent for the variadic model
            inner_epochs : int
                Number of epochs to train for each query size inside the variadic
                model
            callbacks : list
                List of callbacks to be called during optimization
            validation_split : float (range : [0,1])
                Percentage of instances to split off to validate on
            verbose : bool
                Print verbose information
            global_lr : float
                Learning rate of the meta gradient descent (variadic model only)
            global_momentum : float
                Momentum for the meta gradient descent (variadic model only)
            min_bucket_size : int
                Restrict the training to queries of a minimum size
            refit : bool
                If True, create a new model object, otherwise continue fitting the
                existing one if one exists.
            **kwargs :
                Keyword arguments for the fit function
        """
        if optimizer is not None:
            self.optimizer = optimizer
        if isinstance(X, dict):
            if generator is not None:
                self.logger.error("Variadic training does not support generators yet.")
                raise NotImplementedError
            self.is_variadic = True
            decay_rate = global_lr / epochs
            learning_rate = global_lr
            freq = self._bucket_frequencies(X, min_bucket_size=min_bucket_size)
            bucket_ids = np.array(tuple(X.keys()))

            #  Create models which need to be trained
            #  Note, that the models share all their weights, the only
            #  difference is the compute graph constructed for back propagation.
            if not hasattr(self, "models_") or refit:
                self.models_ = self._construct_models(X)

            #  Iterate training
            for epoch in range(epochs):

                self.logger.info(
                    "Epoch: {}, Learning rate: {}".format(epoch, learning_rate)
                )

                # In the spirit of mini-batch SGD we also shuffle the buckets
                # each epoch:
                np.random.shuffle(bucket_ids)
                self.curr_bucket_id = bucket_ids[0]

                w_before = np.array(self.get_weights())

                for bucket_id in bucket_ids:
                    self.curr_bucket_id = bucket_id
                    # Skip query sizes with too few instances:
                    if X[bucket_id].shape[0] < min_bucket_size:
                        continue

                    # self.set_weights(start)
                    x = X[bucket_id]
                    y = Y[bucket_id]

                    # Save weight vector for momentum:
                    w_old = w_before
                    w_before = np.array(self.get_weights())
                    self.models_[bucket_id].fit(
                        x=x,
                        y=y,
                        epochs=inner_epochs,
                        batch_size=self.batch_size,
                        validation_split=validation_split,
                        verbose=verbose,
                        **kwargs,
                    )
                    w_after = np.array(self.get_weights())
                    self.set_weights(
                        w_before
                        + learning_rate * freq[bucket_id] * (w_after - w_before)
                        + global_momentum * (w_before - w_old)
                    )
                learning_rate /= 1 + decay_rate * epoch
        else:
            self.is_variadic = False

            if self.model is None or refit:
                if generator is not None:
                    X, Y = next(iter(generator))

                n_inst, n_objects, n_features = X.shape

                self.model = self.construct_model(n_features, n_objects)
            self.logger.info("Fitting started")
            if generator is None:
                self.model.fit(
                    x=X,
                    y=Y,
                    callbacks=callbacks,
                    epochs=epochs,
                    validation_split=validation_split,
                    batch_size=self.batch_size,
                    verbose=verbose,
                    **kwargs,
                )
            else:
                self.model.fit_generator(
                    generator=generator,
                    callbacks=callbacks,
                    epochs=epochs,
                    verbose=verbose,
                    **kwargs,
                )
            self.logger.info("Fitting complete")

    def construct_model(self, n_features, n_objects):
        """
            Construct the FATE-network architecture using the :class:`DeepSet` to learn the context representation
            :math:`\\mu_{C(x)}` for the given query set/context :math:`Q=C(x)`. We construct an input tensor of query
            set :math:`Q` of size (n_objects, n_features),iterate over it for each object and concatenate the
            context-representation feature tensor of size :math:`\\lvert  \\mu_{C(x)} \\lvert` into a joint layers.
            So, for each object we share the weights in the joint network and the output of this network is used to
            learn the generalized latent utility score :math:`U (x, \\mu_{C(x)})` of each object :math:`x \\in Q`.

            Parameters
            ----------
            n_features: int
                Features of the objects for which the network is constructed
            n_objects: int
                Size of the query sets for which the network is constructed

            Returns
            -------
             model: keras :class:`Model`
                Neural network to learn the FATE utility score

        """
        input_layer = Input(shape=(n_objects, n_features), name="input_node")
        set_repr = self.set_layer(input_layer)
        scores = self.join_input_layers(
            input_layer,
            set_repr,
            n_objects=n_objects,
            n_layers=self.n_hidden_set_layers,
        )
        model = Model(inputs=input_layer, outputs=scores)

        model.compile(
            loss=self.loss_function, optimizer=self.optimizer, metrics=self.metrics
        )
        return model

    def fit(
        self,
        X,
        Y,
        epochs=35,
        inner_epochs=1,
        callbacks=None,
        validation_split=0.1,
        verbose=0,
        global_lr=1.0,
        global_momentum=0.9,
        min_bucket_size=500,
        refit=False,
        **kwargs,
    ):
        """
            Fit a generic preference learning FATE-network model on a provided set of queries.

            The provided queries can be of a fixed size (numpy arrays) or of
            varying sizes in which case dictionaries are expected as input.

            For varying sizes a meta gradient descent is performed across the
            different query sizes.

            Parameters
            ----------
            X : numpy array or dict
                Feature vectors of the objects
                (n_instances, n_objects, n_features) if numpy array or map from n_objects to numpy arrays
            Y : numpy array or dict
                Preferences in form of rankings or choices for given objects
                (n_instances, n_objects) if numpy array or map from n_objects to numpy arrays
            epochs : int
                Number of epochs to run if training for a fixed query size or
                number of epochs of the meta gradient descent for the variadic model
            inner_epochs : int
                Number of epochs to train for each query size inside the variadic
                model
            callbacks : list
                List of callbacks to be called during optimization
            validation_split : float (range : [0,1])
                Percentage of instances to split off to validate on
            verbose : bool
                Print verbose information
            global_lr : float
                Learning rate of the meta gradient descent (variadic model only)
            global_momentum : float
                Momentum for the meta gradient descent (variadic model only)
            min_bucket_size : int
                Restrict the training to queries of a minimum size
            refit : bool
                If True, create a new model object, otherwise continue fitting the
                existing one if one exists.
            **kwargs :
                Keyword arguments for the fit function
        """
        self.random_state_ = check_random_state(self.random_state)
        _n_instances, self.n_objects_fit_, self.n_object_features_fit_ = X.shape
        self._fit(
            X=X,
            Y=Y,
            epochs=epochs,
            inner_epochs=inner_epochs,
            callbacks=callbacks,
            validation_split=validation_split,
            verbose=verbose,
            global_lr=global_lr,
            global_momentum=global_momentum,
            min_bucket_size=min_bucket_size,
            refit=refit,
            **kwargs,
        )

    def fit_generator(
        self,
        generator,
        epochs=35,
        steps_per_epoch=10,
        inner_epochs=1,
        callbacks=None,
        verbose=0,
        global_lr=1.0,
        global_momentum=0.9,
        min_bucket_size=500,
        refit=False,
        **kwargs,
    ):
        """
            Fit a generic object ranking FATE-network on a set of queries provided by
            a generator.

            The provided queries can be of a fixed size (numpy arrays) or of
            varying sizes in which case dictionaries are expected as input.

            For varying sizes a meta gradient descent is performed across the
            different query sizes.

            Parameters
            ----------
            generator :
                A generator or an instance of `Sequence` (:class:`keras.utils.Sequence`) object in order to avoid
                duplicate data when using multiprocessing.
                The output of the generator must be either
                    - a tuple `(inputs, targets)`
                    - a tuple `(inputs, targets, sample_weights)`.
                This tuple (a single output of the generator) makes a single batch.
                Therefore, all arrays in this tuple must have the same length (equal to the size of this batch).
                Different batches may have different sizes.
                For example, the last batch of the epoch is commonly smaller than the others, if the size of the dataset
                is not divisible by the batch size. The generator is expected to loop over its data indefinitely. An
                epoch finishes when `steps_per_epoch` batches have been seen by the model.
            epochs : int
                Number of epochs to run if training for a fixed query size or
                number of epochs of the meta gradient descent for the variadic model
            steps_per_epoch : int
                Number of batches to train per epoch
            inner_epochs : int
                Number of epochs to train for each query size inside the variadic
                model
            callbacks : list
                List of callbacks to be called during optimization
            verbose : bool
                Print verbose information
            global_lr : float
                Learning rate of the meta gradient descent (variadic model only)
            global_momentum : float
                Momentum for the meta gradient descent (variadic model only)
            min_bucket_size : int
                Restrict the training to queries of a minimum size
            refit : bool
                If True, create a new model object, otherwise continue fitting the
                existing one if one exists.
            **kwargs:
                Keyword arguments for the fit function
        """
        self._fit(
            generator=generator,
            epochs=epochs,
            steps_per_epoch=steps_per_epoch,
            inner_epochs=inner_epochs,
            callbacks=callbacks,
            verbose=verbose,
            global_lr=global_lr,
            global_momentum=global_momentum,
            min_bucket_size=min_bucket_size,
            refit=refit,
            **kwargs,
        )

    def _get_context_representation(self, X, kwargs):
        n_objects = X.shape[-2]
        self.logger.info(
            "Test Set instances {} objects {} features {}".format(*X.shape)
        )
        input_layer_scorer = Input(
            shape=(n_objects, self.n_object_features_fit_), name="input_node"
        )
        if self.n_hidden_set_layers >= 1:
            self.set_layer(input_layer_scorer)
            fr = self.set_layer.cached_models[n_objects].predict(X, **kwargs)
            del self.set_layer.cached_models[n_objects]
            X_n = np.empty(
                (fr.shape[0], n_objects, fr.shape[1] + self.n_object_features_fit_),
                dtype="float",
            )
            for i in range(n_objects):
                X_n[:, i] = np.concatenate((X[:, i], fr), axis=1)
            X = np.copy(X_n)
        return X

    def _predict_scores_fixed(self, X, **kwargs):
        """
            Predict the scores for a fixed ranking size.

            Parameters
            ----------
            X : numpy array
                float (n_instances, n_objects, n_features)

            Returns
            -------
            scores : numpy array
                float (n_instances, n_objects)

        """
        # model = self._construct_scoring_model(n_objects)
        X = self._get_context_representation(X, kwargs)
        n_instances, n_objects, n_features = X.shape
        self.logger.info(
            "After applying the set representations features {}".format(n_features)
        )
        input_layer_joint = Input(
            shape=(n_objects, n_features), name="input_joint_model"
        )
        scores = []

        inputs = [create_input_lambda(i)(input_layer_joint) for i in range(n_objects)]

        for i in range(n_objects):
            joint = inputs[i]
            for j in range(self.n_hidden_joint_layers):
                joint = self.joint_layers[j](joint)
            scores.append(self.scorer(joint))
        scores = concatenate(scores, name="final_scores")
        joint_model = Model(inputs=input_layer_joint, outputs=scores)
        predicted_scores = joint_model.predict(X)
        self.logger.info("Done predicting scores")
        return predicted_scores

    def clear_memory(self, n_objects=5, **kwargs):
        """
            Clear the memory, restores the currently fitted model back to prevent memory leaks.

            Parameters
            ----------
            n_objects : int
                float (n_instances, n_objects, n_features)
            **kwargs :
                Keyword arguments for the function
        """
        if self.hash_file is not None:
            self.model.save_weights(self.hash_file)
            K.clear_session()
            sess = tf.Session()
            K.set_session(sess)
            self.optimizer = self.optimizer.from_config(self._optimizer_config)
            self._construct_layers(
                activation=self.activation,
                kernel_initializer=self.kernel_initializer,
                kernel_regularizer=self.kernel_regularizer,
                **self.kwargs,
            )
            self._create_set_layers(
                activation=self.activation,
                kernel_initializer=self.kernel_initializer,
                kernel_regularizer=self.kernel_regularizer,
                **self.kwargs,
            )
            self.model = self.construct_model(self.n_object_features_fit_, n_objects)
            self.model.load_weights(self.hash_file)
        else:
            self.logger.info("Cannot clear the memory")

    def set_tunable_parameters(
        self,
        n_hidden_set_units=32,
        n_hidden_set_layers=2,
        n_hidden_joint_units=32,
        n_hidden_joint_layers=2,
        reg_strength=1e-4,
        learning_rate=1e-3,
        batch_size=128,
        **point,
    ):
        """
            Set tunable parameters of the FATE-network to the values provided.

            Parameters
            ----------
            n_hidden_set_layers : int
                Number of hidden set layers
            n_hidden_set_units : int
                Number of hidden set units in each set layer
            n_hidden_joint_units: int
                Number of hidden joint layers
            n_hidden_joint_layers: int
                Number of hidden units in each joint layer
            reg_strength: float
                Regularization strength of the regularizer function applied to the `kernel` weights matrix
            learning_rate: float
                Learning rate of the stochastic gradient descent algorithm used by the network
            batch_size: int
                Batch size to use during training
            point: dict
                Dictionary containing parameter values which are not tuned for the network
        """
        FATENetworkCore.set_tunable_parameters(
            self,
            n_hidden_joint_units=n_hidden_joint_units,
            n_hidden_joint_layers=n_hidden_joint_layers,
            reg_strength=reg_strength,
            learning_rate=learning_rate,
            batch_size=batch_size,
            **point,
        )

        self.n_hidden_set_units = n_hidden_set_units
        self.n_hidden_set_layers = n_hidden_set_layers

        self._create_set_layers(
            activation=self.activation,
            kernel_initializer=self.kernel_initializer,
            kernel_regularizer=self.kernel_regularizer,
            **self.kwargs,
        )
        if hasattr(self, "model"):
            self.model = None
        if hasattr(self, "models"):
            self.models = None



try:
    from theano import tensor as tt
except ImportError:
    from csrank.util import MissingExtraError

    raise MissingExtraError("theano", "probabilistic")


def replace_inf_theano(x):
    if tt.any(tt.isinf(x)):
        x = tt.switch(tt.isinf(x), 2e300, x)
        x = tt.switch(tt.isnan(x), 2e300, x)
    return x


def replace_nan_theano(p):
    if tt.any(tt.isnan(p)):
        p = tt.switch(tt.isnan(p), 1.0, p)
        p = tt.switch(tt.isinf(p), 1.0, p)
        p = normalize(p)
    return p


def logsumexp(x, axis=1):
    x_max = tt.max(x, axis=axis, keepdims=True)
    return tt.log(tt.sum(tt.exp(x - x_max), axis=axis, keepdims=True)) + x_max


def softmax(x, axis=1):
    """
        Take softmax for the given two dimensional theano array.
        :param axis: The axis around which the softmax is applied
        :param x: array-like, shape (n_samples, n_objects)
        :return: softmax taken around the axis=1
    """
    lse = logsumexp(x, axis=axis)
    return tt.exp(x - lse)


def sigmoid(x):
    x = 1 / (1 + tt.exp(-1 * x))
    return x


def normalize(x, axis=1):
    """
        Normalize the given two dimensional theano array around the row.
        :param axis: The axis around which the norm is applied
        :param x: theano or numpy array-like, shape (n_samples, n_objects)
        :return: normalize the array around the axis=1
    """
    return x / tt.sum(x, axis=axis, keepdims=True)



from abc import ABCMeta
from abc import abstractmethod

from csrank.tunable import Tunable


class Learner(Tunable, metaclass=ABCMeta):
    @abstractmethod
    def fit(self, X, Y, **kwargs):
        """
            Fit the preference learning algorithm on the provided set of queries X and preferences Y of those objects.
            The provided queries and corresponding preferences are of a fixed size (numpy arrays).

            Parameters
            ----------
            X : array-like, shape (n_samples, n_objects, n_features)
                Feature vectors of the objects
            Y : array-like, shape (n_samples, n_objects)
                Preferences of the objects in form of rankings or choices
        """
        raise NotImplementedError

    @abstractmethod
    def _predict_scores_fixed(self, X, **kwargs):
        """
            Predict the scores for a given collection of sets of objects of same size.

            Parameters
            ----------
            X : array-like, shape (n_samples, n_objects, n_features)


            Returns
            -------
            Y : array-like, shape (n_samples, n_objects)
                Returns the scores of each of the objects for each of the samples.
        """
        raise NotImplementedError

    def predict_for_scores(self, scores, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def predict_scores(self, X, **kwargs):
        """
            Predict the utility scores for each object in the collection of set of objects called a query set.

            Parameters
            ----------
            X : dict or numpy array
                Dictionary with a mapping from query set size to numpy arrays or a single numpy array of size:
                (n_instances, n_objects, n_features)

            Returns
            -------
            Y : dict or numpy array
                Dictionary with a mapping from query set size to numpy arrays or a single numpy array of size:
                (n_instances, n_objects)
        """
        self.logger.info("Predicting scores")

        if isinstance(X, dict):
            scores = dict()
            for ranking_size, x in X.items():
                n_instances, n_objects, n_features = x.shape
                if "clear_memory" in dir(self):
                    self.clear_memory(
                        n_instances=n_instances,
                        n_objects=n_objects,
                        n_features=n_features,
                    )
                scores[ranking_size] = self._predict_scores_fixed(x, **kwargs)

        else:
            scores = self._predict_scores_fixed(X, **kwargs)
        return scores

    @abstractmethod
    def predict(self, X, **kwargs):
        """
            Predict preferences in the form of rankings or choices for a given collection of sets of objects called
            a query set using the function :meth:`.predict_for_scores`.

            Parameters
            ----------
            X : dict or numpy array
                Dictionary with a mapping from the query set size to numpy arrays or a single numpy array of size:
                (n_instances, n_objects, n_features)


            Returns
            -------
            Y : dict or numpy array
                Dictionary with a mapping from the query set size to numpy arrays or a single numpy array containing
                predicted preferences of size:
                (n_instances, n_objects)
        """
        self.logger.debug("Predicting started")

        scores = self.predict_scores(X, **kwargs)
        self.logger.debug("Predicting scores complete")

        return self.predict_for_scores(scores, **kwargs)

    def __call__(self, X, *args, **kwargs):
        """
        Predict preferences in the form of rankings or choices for a given collection of sets of objects.

        Parameters
        ----------
        X : dict or numpy array
            Dictionary with a mapping from ranking size to numpy arrays
            or a single numpy array of size:
            (n_instances, n_objects, n_features)


        Returns
        -------
        Y : Dictionary with a mapping from ranking size to numpy arrays
            or a single numpy array of size:
            (n_instances, n_objects)
            Predicted ranking
        """
        return self.predict(X, **kwargs)

    @classmethod
    def __subclasshook__(cls, C):
        if cls is Learner:
            has_fit = any("fit" in B.__dict__ for B in C.__mro__)
            has_predict = any("predict" in B.__dict__ for B in C.__mro__)
            has_scores = any("predict_scores" in B.__dict__ for B in C.__mro__)
            has_scores_fixed = any(
                "_predict_scores_fixed" in B.__dict__ for B in C.__mro__
            )
            has_predict_for_scores = any(
                "predict_for_scores" in B.__dict__ for B in C.__mro__
            )
            if (
                has_fit
                and has_predict
                and has_scores
                and has_scores_fixed
                and has_predict_for_scores
            ):
                return True
        return NotImplemented



import os

from keras.optimizers import SGD
import numpy as np
from pymc3.variational.callbacks import CheckParametersConvergence
import pytest
import tensorflow as tf

from csrank.constants import CMPNET_DC
from csrank.constants import FATE_DC
from csrank.constants import FATELINEAR_DC
from csrank.constants import FETA_DC
from csrank.constants import FETALINEAR_DC
from csrank.constants import GEV
from csrank.constants import MLM
from csrank.constants import MNL
from csrank.constants import NLM
from csrank.constants import PCL
from csrank.constants import RANKNET_DC
from csrank.constants import RANKSVM_DC
from csrank.dataset_reader.discretechoice.util import convert_to_label_encoding
from csrank.discretechoice import *
from csrank.metrics_np import categorical_accuracy_np
from csrank.metrics_np import subset_01_loss
from csrank.metrics_np import topk_categorical_accuracy_np
from csrank.tests.test_ranking import check_params_tunable
from csrank.util import metrics_on_predictions

metrics = {
    "CategoricalAccuracy": categorical_accuracy_np,
    "CategoricalTopK2": topk_categorical_accuracy_np(k=2),
}
optimizer = SGD(lr=1e-3, momentum=0.9, nesterov=True)


def get_vals(values=[1.0, 1.0]):
    return dict(zip(metrics.keys(), values))


discrete_choice_functions = {
    FETA_DC: (
        FETADiscreteChoiceFunction,
        {"n_hidden": 1, "optimizer": optimizer},
        get_vals([0.978, 1.0]),
    ),
    RANKNET_DC: (
        RankNetDiscreteChoiceFunction,
        {"optimizer": optimizer},
        get_vals([0.97, 0.996]),
    ),
    CMPNET_DC: (
        CmpNetDiscreteChoiceFunction,
        {"optimizer": optimizer},
        get_vals([0.994, 1.0]),
    ),
    FATE_DC: (
        FATEDiscreteChoiceFunction,
        {
            "n_hidden_joint_layers": 1,
            "n_hidden_set_layers": 1,
            "n_hidden_joint_units": 5,
            "n_hidden_set_units": 5,
            "optimizer": optimizer,
        },
        get_vals([0.95, 0.998]),
    ),
    FATELINEAR_DC: (
        FATELinearDiscreteChoiceFunction,
        {"n_hidden_set_units": 1, "learning_rate": 5e-3, "batch_size": 32},
        get_vals([0.74, 0.934]),
    ),
    FETALINEAR_DC: (FETALinearDiscreteChoiceFunction, {}, get_vals([0.976, 0.9998])),
    MNL: (MultinomialLogitModel, {}, get_vals([0.998, 1.0])),
    NLM: (NestedLogitModel, {}, get_vals()),
    PCL: (PairedCombinatorialLogit, {}, get_vals()),
    GEV: (GeneralizedNestedLogitModel, {}, get_vals()),
    MLM: (MixedLogitModel, {}, get_vals()),
    RANKSVM_DC: (PairwiseSVMDiscreteChoiceFunction, {}, get_vals([0.982, 0.982])),
}


@pytest.fixture(scope="module")
def trivial_discrete_choice_problem():
    random_state = np.random.RandomState(42)
    x = random_state.randn(500, 5, 2)
    w = random_state.rand(2)
    y_true = np.argmax(np.dot(x, w), axis=1)
    y_true = convert_to_label_encoding(y_true, 5)
    return x, y_true


@pytest.mark.parametrize("name", list(discrete_choice_functions.keys()))
def test_discrete_choice_function_fixed(trivial_discrete_choice_problem, name):
    tf.set_random_seed(0)
    os.environ["KERAS_BACKEND"] = "tensorflow"
    np.random.seed(123)
    x, y = trivial_discrete_choice_problem
    choice_function = discrete_choice_functions[name][0]
    params, accuracies = (
        discrete_choice_functions[name][1],
        discrete_choice_functions[name][2],
    )
    learner = choice_function(**params)
    if name in [MNL, NLM, GEV, PCL, MLM]:
        learner.fit(
            x,
            y,
            vi_params={
                "n": 100,
                "method": "advi",
                "callbacks": [CheckParametersConvergence()],
            },
        )
    elif "linear" in name:
        learner.fit(x, y, epochs=10, validation_split=0, verbose=False)
    else:
        learner.fit(x, y, epochs=100, validation_split=0, verbose=False)
    s_pred = learner.predict_scores(x)
    y_pred = learner.predict_for_scores(s_pred)
    y_pred_2 = learner.predict(x)
    rtol = 1e-2
    atol = 5e-2
    assert np.isclose(
        0.0, subset_01_loss(y_pred, y_pred_2), rtol=rtol, atol=atol, equal_nan=False
    )
    for key, value in accuracies.items():
        metric = metrics[key]
        if metric in metrics_on_predictions:
            pred_loss = metric(y, y_pred)
        else:
            pred_loss = metric(y, s_pred)
        assert np.isclose(value, pred_loss, rtol=rtol, atol=atol, equal_nan=False)

    params = {
        "n_hidden": 20,
        "n_units": 20,
        "n_hidden_set_units": 2,
        "n_hidden_set_layers": 10,
        "n_hidden_joint_units": 2,
        "n_hidden_joint_layers": 10,
        "reg_strength": 1e-3,
        "learning_rate": 1e-1,
        "batch_size": 32,
        "alpha": 0.5,
        "l1_ratio": 0.7,
        "tol": 1e-2,
        "C": 10,
        "n_mixtures": 10,
        "n_nests": 5,
        "regularization": "l2",
    }
    learner.set_tunable_parameters(**params)
    check_params_tunable(learner, params, rtol, atol)



import numpy as np


def sub_sampling_choices_from_relevance(Xt, Yt, n_objects=5, offset=2):
    X_train = []
    Y_train = []
    bucket_size = int(Xt.shape[1] / n_objects) + offset
    rs = np.random.RandomState(42 + n_objects)
    for x, y in zip(Xt, Yt):
        choices = np.append(np.where(y == 1)[0], np.where(y == 2)[0])
        zeros = np.delete(np.arange(0, Xt.shape[1]), choices)
        y[choices] = 1
        if len(zeros) != 0 and len(choices) >= 1:
            c_max = n_objects if n_objects < len(choices) else len(choices) + 1
            sizes = rs.choice(np.arange(1, c_max), size=bucket_size)
            if len(zeros) > n_objects - 1:
                idx = np.array(
                    [
                        np.append(
                            rs.choice(choices, size=size, replace=False),
                            rs.choice(zeros, size=(n_objects - size), replace=False),
                        )
                        for size in sizes
                    ]
                )
            else:
                idx = np.array(
                    [
                        np.append(
                            rs.choice(choices, size=size, replace=False),
                            rs.choice(zeros, size=(n_objects - size), replace=True),
                        )
                        for size in sizes
                    ]
                )
            for i in idx:
                np.random.shuffle(i)
            if len(X_train) == 0:
                X_train = x[idx]
                Y_train = y[idx]
            else:
                Y_train = np.concatenate([Y_train, y[idx]], axis=0)
                X_train = np.concatenate([X_train, x[idx]], axis=0)
    return X_train, Y_train



from abc import ABCMeta

from csrank.constants import OBJECT_RANKING
from csrank.numpy_util import scores_to_rankings

__all__ = ["ObjectRanker"]


class ObjectRanker(metaclass=ABCMeta):
    @property
    def learning_problem(self):
        return OBJECT_RANKING

    def predict_for_scores(self, scores, **kwargs):
        """
            The permutation vector :math:`\\pi` represents the ranking amongst the objects in :math:`Q`, such that
            :math:`\\pi(k)` is the position of the :math:`k`-th object :math:`x_k`, and :math:`\\pi^{-1}(k)` is the index
            of the object on position :math:`k`. Predict rankings for the scores for a given collection of sets of
            objects (query sets).

            Parameters
            ----------
            scores : dict or numpy array
                Dictionary with a mapping from query set size to numpy arrays or a single numpy array of size containing
                scores of each object of size: (n_instances, n_objects)

            Returns
            -------
            Y : dict or numpy array
                Dictionary with a mapping from objects size to numpy arrays or a single numpy array containing
                predicted rankings of size: (n_samples, n_objects)
        """

        if isinstance(scores, dict):
            result = dict()
            for n, score in scores.items():
                rankings = scores_to_rankings(score)
                result[n] = rankings
        else:
            result = scores_to_rankings(scores)
        return result



import logging

from csrank.core.pairwise_svm import PairwiseSVM
from csrank.objectranking.object_ranker import ObjectRanker
from ..dataset_reader.objectranking.util import generate_complete_pairwise_dataset

__all__ = ["RankSVM"]


class RankSVM(ObjectRanker, PairwiseSVM):
    def __init__(
        self,
        C=1.0,
        tol=1e-4,
        normalize=True,
        fit_intercept=True,
        random_state=None,
        **kwargs,
    ):
        """
            Create an instance of the :class:`PairwiseSVM` model for learning a object ranking function.
            It learns a linear deterministic utility function of the form :math:`U(x) = w \\cdot x`, where :math:`w` is
            the weight vector. It is estimated using *pairwise preferences* generated from the rankings.
            The ranking for the given query set :math:`Q` is defined as:

            .. math::

                ρ(Q)  = \\operatorname{argsort}_{x \\in Q}  \\; U(x)

            Parameters
            ----------
            C : float, optional
                Penalty parameter of the error term
            tol : float, optional
                Optimization tolerance
            normalize : bool, optional
                If True, the data will be normalized before fitting.
            fit_intercept : bool, optional
                If True, the linear model will also fit an intercept.
            random_state : int, RandomState instance or None, optional
                Seed of the pseudorandom generator or a RandomState instance
            **kwargs
                Keyword arguments for the algorithms

            References
            ----------
                [1] Joachims, T. (2002, July). "Optimizing search engines using clickthrough data.", Proceedings of the eighth ACM SIGKDD international conference on Knowledge discovery and data mining (pp. 133-142). ACM.
        """
        super().__init__(
            C=C,
            tol=tol,
            normalize=normalize,
            fit_intercept=fit_intercept,
            random_state=random_state,
            **kwargs,
        )
        self.logger = logging.getLogger(RankSVM.__name__)
        self.logger.info("Initializing network")

    def fit(self, X, Y, **kwargs):
        _n_instances, self.n_objects_fit_, self.n_object_features_fit_ = X.shape
        super().fit(X, Y, **kwargs)

    def _convert_instances_(self, X, Y):
        self.logger.debug("Creating the Dataset")
        (
            x_train,
            garbage,
            garbage,
            garbage,
            y_single,
        ) = generate_complete_pairwise_dataset(X, Y)
        del garbage
        assert x_train.shape[1] == self.n_object_features_fit_
        self.logger.debug(
            "Finished the Dataset with instances {}".format(x_train.shape[0])
        )
        return x_train, y_single

    def _predict_scores_fixed(self, X, **kwargs):
        return super()._predict_scores_fixed(X, **kwargs)

    def predict_scores(self, X, **kwargs):
        return super().predict_scores(X, **kwargs)

    def predict_for_scores(self, scores, **kwargs):
        return ObjectRanker.predict_for_scores(self, scores, **kwargs)

    def predict(self, X, **kwargs):
        return super().predict(X, **kwargs)

    def set_tunable_parameters(self, C=1.0, tol=1e-4, **point):
        super().set_tunable_parameters(C=C, tol=tol, **point)



LABEL_RANKING = "label_ranking"
DYAD_RANKING = "dyad_ranking"
OBJECT_RANKING = "object_ranking"
DISCRETE_CHOICE = "discrete_choice"
CHOICE_FUNCTION = "choice_function"
EXCEPTION_OUTPUT_FEATURES_INSTANCES = (
    "Number of instances inconsistent for {} dataset! output array:{} instances "
    "and features arrays:{} instances"
)
EXCEPTION_OBJECT_ARRAY_SHAPE = (
    "Invalid shape for {} dataset objects features array! shape is: {}"
)
EXCEPTION_CONTEXT_ARRAY_SHAPE = (
    "Invalid shape for {} dataset context features array! shape is: {}"
)
EXCEPTION_UNWANTED_CONTEXT_FEATURES = "Unwanted extra context features in {} dataset"
EXCEPTION_RANKINGS_FEATURES_NO_OF_OBJECTS = (
    "Number of objects inconsistent! in rankings:{} objects "
    "and features array: {} objects"
)
EXCEPTION_RANKINGS = "Unwanted rankings in {} dataset"
EXCEPTION_SET_INCLUSION = "Choice Set inclusion/exclusion binary code not present for all the objects in the set."
allowed_dense_kwargs = [
    "input_shape",
    "batch_input_shape",
    "batch_size",
    "dtype",
    "name",
    "trainable",
    "weights",
    "input_dtype",
    "activation",
    "use_bias",
    "kernel_initializer",
    "bias_initializer",
    "kernel_regularizer",
    "bias_regularizer",
    "activity_regularizer",
    "kernel_constraint",
    "bias_constraint",
]

RANKSVM = "ranksvm"

ERR = "err"
CMPNET = "cmpnet"
RANKNET = "ranknet"
FETA_RANKER = "feta_ranker"
FATE_RANKER = "fate_ranker"
LISTNET = "listnet"
FATELINEAR_RANKER = "fatelinear_ranker"
FETALINEAR_RANKER = "fetalinear_ranker"
RANDOM_RANKER = "random_ranker"

FETA_CHOICE = "feta_choice"
FETALINEAR_CHOICE = "fetalinear_choice"
FATE_CHOICE = "fate_choice"
FATELINEAR_CHOICE = "fatelinear_choice"
RANKNET_CHOICE = "ranknet_choice"
CMPNET_CHOICE = "cmpnet_choice"
RANKSVM_CHOICE = "ranksvm_choice"
GLM_CHOICE = "glm_choice"
RANDOM_CHOICE = "random_choice"

FETA_DC = "feta_dc"
FATE_DC = "fate_dc"
FATELINEAR_DC = "fatelinear_dc"
FETALINEAR_DC = "fetalinear_dc"
RANDOM_DC = "random_dc"

RANKNET_DC = "ranknet_dc"
CMPNET_DC = "cmpnet_dc"
MNL = "multinomial_logit_model"
NLM = "nested_logit_model"
GEV = "generalized_extreme_value"
PCL = "paired_combinatorial_logit"
RANKSVM_DC = "ranksvm_dc"
MLM = "mixed_logit_model"



import logging

from keras.layers import Activation
from keras.layers import BatchNormalization
from keras.layers import Dense
from keras.layers import Input
from keras.layers import Lambda
from keras.layers.merge import average
from keras.models import Model

__all__ = ["NormalizedDense", "DeepSet", "create_input_lambda"]


class NormalizedDense(object):
    """Stop training when a monitored quantity has stopped improving.
        # Arguments
        units: Positive integer, dimensionality of the output space.
        activation: Activation function to use
            (see [activations](../activations.md)).
            If you don't specify anything, no activation is applied
            (ie. "relu:).
        normalize_before_activation: True if normalize the inputs before applying the activation.
        False if activation is applied before Bach Normalization
        """

    def __init__(
        self, units, activation="relu", normalize_before_activation=False, **kwd
    ):
        self.dense = Dense(units, activation="linear", **kwd)
        self.activation = Activation(activation=activation)
        self.batchnorm = BatchNormalization()
        self.norm_layer = None
        self.normalize_before_activation = normalize_before_activation

    def __call__(self, x):
        if self.normalize_before_activation:
            return self.activation(self.batchnorm(self.dense(x)))
        else:
            return self.batchnorm(self.activation(self.dense(x)))

    def get_weights(self):
        w_b = self.batchnorm.get_weights()
        w_d = self.dense.get_weights()
        return w_b, w_d

    def set_weights(self, weights):
        w_b, w_d = weights
        self.batchnorm.set_weights(w_b)
        self.dense.set_weights(w_d)


class DeepSet(object):
    """Deep layer for learning representations for sets of objects.

    Parameters
    ----------
    units : int
        Number of units in each representation layer

    layers : int
        Number of layers to use for learning the representation

    activation : string, optional (default='selu')
        Activation function to use in each unit

    kernel_initializer : string, optional (default='lecun_normal')
        Initializer for the weight matrix

    input_shape : array_like
        Should provide (n_objects, n_features) (DEPRECATED)

    Attributes
    ----------
    model : Keras model
        Representing the complete deep set layer

    set_mapping_layers : list
        List of densely connected hidden layers
    """

    def __init__(
        self,
        units,
        layers=1,
        activation="selu",
        kernel_initializer="lecun_normal",
        kernel_regularizer=None,
        input_shape=None,
        **kwargs,
    ):
        self.logger = logging.getLogger("DeepSets")
        self.n_units = units
        if input_shape is not None:
            self.logger.warning(
                "input_shape is deprecated, since the number "
                "of objects is now inferred"
            )
            self.n_features = input_shape[1]
        self.n_layers = layers
        self.activation = activation
        self.kernel_initializer = kernel_initializer
        self.kernel_regularizer = kernel_regularizer

        self.cached_models = dict()
        self._construct_layers(
            kernel_initializer=kernel_initializer,
            kernel_regularizer=kernel_regularizer,
            activation=activation,
            **kwargs,
        )

    def _construct_layers(self, **kwargs):
        # Create set representation layers:
        self.set_mapping_layers = []
        for i in range(self.n_layers):
            self.set_mapping_layers.append(
                Dense(self.n_units, name="set_layer_{}".format(i), **kwargs)
            )

    def _create_model(self, shape):
        n_objects, n_features = shape[1].value, shape[2].value
        if hasattr(self, "n_features"):
            if self.n_features != n_features:
                self.logger.error("Number of features is not consistent.")
        input_layer = Input(shape=(n_objects, n_features))
        inputs = [create_input_lambda(i)(input_layer) for i in range(n_objects)]

        # Connect input tensors with set mapping layer:
        set_mappings = []
        for i in range(n_objects):
            curr = inputs[i]
            for j in range(len(self.set_mapping_layers)):
                curr = self.set_mapping_layers[j](curr)
            set_mappings.append((i, curr))

        # TODO: is feature_repr used outside?
        feature_repr = average([x for (j, x) in set_mappings])

        self.cached_models[n_objects] = Model(inputs=input_layer, outputs=feature_repr)

    def __call__(self, x):
        shape = x.shape
        n_objects = shape[1].value
        if n_objects not in self.cached_models:
            self._create_model(shape)
        return self.cached_models[n_objects](x)

    def get_weights(self):
        w_set = [x.get_weights() for x in self.set_mapping_layers]
        return w_set

    def set_weights(self, weights):
        for i, layer in enumerate(self.set_mapping_layers):
            layer.set_weights(weights[i])


def create_input_lambda(i):
    """Extracts off an object tensor from an input tensor"""
    return Lambda(lambda x: x[:, i])



from itertools import combinations
import logging

import numpy as np

from csrank.numpy_util import ranking_ordering_conversion

__all__ = [
    "generate_complete_pairwise_dataset",
    "complete_linear_regression_dataset",
    "complete_linear_regression_dataset",
    "sub_sampling_rankings",
]


def generate_pairwise_instances(features):
    pairs = np.array(list(combinations(features, 2)))

    n_pairs = len(pairs)
    neg_indices = np.arange(0, n_pairs, 2)

    a, b = np.copy(pairs[neg_indices, 0]), np.copy(pairs[neg_indices, 1])
    pairs[neg_indices, 1] = a
    pairs[neg_indices, 0] = b

    X1 = pairs[:, 0]
    X2 = pairs[:, 1]
    Y_double = np.ones([n_pairs, 1]) * np.array([1, 0])
    Y_single = np.repeat(1, n_pairs)

    Y_double[neg_indices] = [0, 1]
    Y_single[neg_indices] = 0
    return X1, X2, Y_double, Y_single


def generate_complete_pairwise_dataset(X, Y):
    """
        Generates the pairiwse preference data from the given rankings.The ranking amongst the objects in a query set
        :math:`Q = \\{x_1, x_2, x_3\\}` is represented by :math:`\\pi = (2,1,3)`, such that :math:`\\pi(2)=1` is the position of the :math:`x_2`.
        One can extract the following *pairwise preferences* :math:`x_2 \\succ x_1, x_2 \\succ x_3 and x_1 \\succ x_3`.
        This function generates pairwise preferences which can be used to learn different :class:`ObjectRanker` as:
            1. :class:`RankNet`
            2. :class:`CmpNet`
            3. :class:`RankSVM`

        Parameters
        ----------
        X : numpy array (n_instances, n_objects, n_features)
            Feature vectors of the objects
        Y : numpy array (n_instances, n_objects)
            Rankings of the given objects

        Returns
        -------
        x_train: array-like, shape (n_samples, n_features)
            The difference vector between the two objects with pairwise preference :math:`x_2 \\succ x_1`, i.e.
            :math:`x_2-x_1` with n_samples=:math:`n_{instances} \\cdot (n_{objects} \\choose 2)`. :class:`RankSVM` uses
            it as input.
        x_train1: array-like, shape (n_samples, n_features)
            The first object :math:`x_2` of the pairwise preference :math:`x_2 \\succ x_1`, with
            n_samples=:math:`n_{instances} \\cdot (n_{objects} \\choose 2)`. :class:`RankNet` and :class:`CmpNet` uses it
            as input.
        x_train2: array-like, shape (n_samples, n_features)
            The second object :math:`x_1` of the pairwise preference :math:`x_2 \\succ x_1`, with
            n_samples=:math:`n_{instances} \\cdot (n_{objects} \\choose 2)`. :class:`RankNet` and :class:`CmpNet` uses it
            as input.
        y_double: array-like, shape (n_samples, 2)
            The preference :math:`x_2 \\succ x_1` between the objects :math:`x_2` and :math:`x_1`, i.e. (1,0)
            with n_samples=:math:`n_{instances} \\cdot (n_{objects} \\choose 2)`. :class:`CmpNet` uses it
            as input.
        y_double: array-like, shape (n_samples)
            The preference :math:`x_2 \\succ x_1` between the objects :math:`x_2` and :math:`x_1`, i.e. 1
            with n_samples=:math:`n_{instances} \\cdot (n_{objects} \\choose 2)`.:class:`RankNet` and :class:`RankSVM`
            uses it as output.
    """
    try:
        n_instances, n_objects, n_features = X.shape
        Y = Y.astype(int)
        Y -= np.min(Y)
        orderings = ranking_ordering_conversion(Y)
        x_sorted = [X[i, orderings[i], :] for i in range(n_instances)]
        del orderings
    except ValueError:
        # TODO Add the code to change the rankings to orderings and sort X according to that
        logger = logging.getLogger("generate_complete_pairwise_dataset")
        logger.error("Value Error: {}, {} ".format(X[0], Y[0]))
        x_sorted = X
    del Y
    y_double = []
    x_train1 = []
    x_train2 = []
    y_single = []
    for features in x_sorted:
        x1, x2, y1, y2 = generate_pairwise_instances(features)
        x_train1.extend(x1)
        x_train2.extend(x2)
        y_double.extend(y1)
        y_single.extend(y2)
    x_train1 = np.array(x_train1)
    x_train2 = np.array(x_train2)
    x_train = x_train1 - x_train2
    y_double = np.array(y_double)
    y_single = np.array(y_single)
    return x_train, x_train1, x_train2, y_double, y_single


def complete_linear_regression_dataset(X, rankings):
    X1 = []
    Y_single = []
    for features, rank in zip(X, rankings):
        X1.extend(features)
        norm_ranks = rank / np.max(rank, axis=0)
        Y_single.extend(norm_ranks)
    X1 = np.array(X1)
    Y_single = np.array(Y_single)
    return X1, Y_single


def sub_sampling_rankings(Xt, Yt, n_objects=5):
    bucket_size = int(Xt.shape[1] / n_objects)
    X_train = []
    Y_train = []
    for i in range(bucket_size):
        X = np.copy(Xt)
        Y = np.copy(Yt)
        rs = np.random.RandomState(42 + i)
        idx = rs.randint(bucket_size, size=(len(X), n_objects))
        idx += np.arange(start=0, stop=X.shape[1], step=bucket_size)[:n_objects]
        X = X[np.arange(X.shape[0])[:, None], idx]
        Y = Y[np.arange(X.shape[0])[:, None], idx]
        tmp_sort = Y.argsort(axis=-1)
        Y = np.empty_like(Y)
        Y[np.arange(len(X))[:, None], tmp_sort] = np.arange(n_objects)
        if len(X_train) == 0:
            X_train = X
            Y_train = Y
        else:
            Y_train = np.concatenate([Y_train, Y], axis=0)
            X_train = np.concatenate([X_train, X], axis=0)
    return X_train, Y_train



import collections
import glob
from itertools import combinations
from itertools import product
import logging
import os

import numpy as np
from sklearn.preprocessing import StandardScaler
from sklearn.utils import check_random_state

from csrank.constants import OBJECT_RANKING
from csrank.numpy_util import scores_to_rankings
from ..dataset_reader import DatasetReader
from ..util import distance_metric_multilabel
from ..util import get_key_for_indices
from ..util import get_similarity_matrix

try:
    import h5py
except ImportError:
    from csrank.util import MissingExtraError

    raise MissingExtraError("h5py", "data")

try:
    import pandas as pd
except ImportError:
    from csrank.util import MissingExtraError

    raise MissingExtraError("pandas", "data")


class ImageDatasetReader(DatasetReader):
    def __init__(
        self,
        n_train_instances=10000,
        n_test_instances=10000,
        n_objects=5,
        random_state=None,
        **kwargs,
    ):
        super(ImageDatasetReader, self).__init__(
            learning_problem=OBJECT_RANKING, dataset_folder="image_dataset", **kwargs
        )
        TRAIN_LABEL = "gt_label_train"
        VAL_LABEL = "gt_label_val"
        # TEST_LABEL = 'gt_label_test'

        # self.test_file = os.path.join(dirname, DATASET_FOLDER, "test.hd5")
        # self.labels_test_files = glob.glob(os.path.join(dirname, DATASET_FOLDER, TEST_LABEL, "*.txt"))
        # self.labels_test_files.sort()
        self.logger = logging.getLogger(ImageDatasetReader.__name__)

        self.train_file = os.path.join(self.dirname, "train.hd5")
        self.similarity_matrix_train_file = os.path.join(
            self.dirname, "train_similarity_matrix.csv"
        )
        self.labels_train_files = glob.glob(
            os.path.join(self.dirname, TRAIN_LABEL, "*.txt")
        )
        self.labels_train_files.sort()

        self.test_file = os.path.join(self.dirname, "val.hd5")
        self.similarity_matrix_test_file = os.path.join(
            self.dirname, "val_similarity_matrix.csv"
        )
        self.labels_test_files = glob.glob(
            os.path.join(self.dirname, VAL_LABEL, "*.txt")
        )
        self.labels_test_files.sort()

        self.n_test_instances = n_test_instances
        self.n_train_instances = n_train_instances
        self.n_objects = n_objects
        self.n_features = 2048
        self.label_names = []
        for file in self.labels_test_files:
            self.label_names.append(os.path.basename(file).split(".")[0].split("_")[0])

        self.logger.info("Image Features Train File {}".format(self.train_file))
        self.logger.info("Image Features Test File {}".format(self.test_file))
        self.logger.info("Image Labels Train File {}".format(self.labels_train_files))
        self.logger.info("Image Labels Test File {}".format(self.labels_test_files))
        self.logger.info("Image Labels Names {}".format(self.label_names))
        self.random_state = check_random_state(random_state)
        if not (
            os.path.isfile(self.similarity_matrix_test_file)
            and os.path.isfile(self.similarity_matrix_train_file)
        ):
            self.__load_dataset__()

    def __load_dataset__(self):
        self.create_multilabel_dataset(datatype="train")
        self.create_multilabel_dataset(datatype="test")

    def splitter(self, iter):
        for i in iter:
            self.X, self.Y = X_train, Y_train = self.make_similarity_based_dataset(
                datatype="train", seed=10 * i + 32
            )
            self.__check_dataset_validity__()

            self.X, self.Y = X_test, Y_test = self.make_similarity_based_dataset(
                datatype="test", seed=10 * i + 32
            )
            self.__check_dataset_validity__()

        yield X_train, Y_train, X_test, Y_test

    def get_dataset_dictionaries(self):
        pass

    def get_train_test_datasets(self, n_datasets=5):
        splits = np.array(n_datasets)
        return self.splitter(splits)

    def get_single_train_test_split(self):
        seed = self.random_state.randint(2 ** 32, dtype="uint32")
        self.X, self.Y = X_train, Y_train = self.make_similarity_based_dataset(
            datatype="train", seed=seed
        )
        self.__check_dataset_validity__()

        self.X, self.Y = X_test, Y_test = self.make_similarity_based_dataset(
            datatype="test", seed=seed + 1
        )
        self.__check_dataset_validity__()
        return X_train, Y_train, X_test, Y_test

    def make_similarity_based_dataset(self, datatype="train", seed=42):
        """Picks a random subset of objects, determines the medoid and ranks the objects
        based on the distance to the medoid.

        The medoid is also included in the ordering."""
        random_state = np.random.RandomState(seed=seed)
        if datatype == "train":
            image_features = self.image_features_train
            n_instances = self.n_train_instances
            similarity_matrix_file = self.similarity_matrix_train_file
        elif datatype == "test":
            image_features = self.image_features_test
            n_instances = self.n_test_instances
            similarity_matrix_file = self.similarity_matrix_test_file

        X = np.empty((n_instances, self.n_objects, self.n_features), dtype=float)
        similarity_scores = np.empty((n_instances, self.n_objects), dtype=float)
        similarity_matrix_lin_list = get_similarity_matrix(similarity_matrix_file)

        for i in range(n_instances):
            subset = random_state.choice(
                image_features.shape[0], size=self.n_objects, replace=False
            )
            X[i] = image_features[subset]
            query = random_state.choice(self.n_objects, size=1)
            one_row = [
                similarity_matrix_lin_list[get_key_for_indices(i, j)]
                for i, j in product(subset[query], subset)
            ]
            similarity_scores[i] = np.array(one_row)

        Y = scores_to_rankings(similarity_scores)
        for i, x in enumerate(X):
            x = StandardScaler().fit_transform(x)
            X[i] = x
        return X, Y

    def create_multilabel_dataset(self, datatype="train"):
        if datatype == "train":
            image_file_name = self.train_file
            label_file_names = self.labels_train_files
            similarity_matrix_file = self.similarity_matrix_train_file
        elif datatype == "test":
            image_file_name = self.test_file
            label_file_names = self.labels_test_files
            similarity_matrix_file = self.similarity_matrix_test_file

        image_file = h5py.File(image_file_name, "r")
        names = np.array(image_file.get("name")[()], dtype=str)
        image_features = np.array(image_file.get("feature"))

        names = [str.replace(w, "b", "") for w in names]
        names = [str.replace(w, "\n", "") for w in names]
        label_vectors_dictionary = {key: [] for key in names}

        label_keys = []
        for file in label_file_names:
            lines = np.array([line.rstrip("\n") for line in open(file)])
            for line in lines:
                label_keys.append(line.split(" ")[0])
                if line.split(" ")[0] in label_vectors_dictionary:
                    label_vectors_dictionary[line.split(" ")[0]].append(
                        int(line.split(" ")[-1])
                    )
        label_vectors_dictionary = collections.OrderedDict(
            sorted(label_vectors_dictionary.items())
        )
        label_vectors = []
        for key, value in label_vectors_dictionary.items():
            label_vectors.append(value)
        label_keys = list(set(label_keys))
        label_keys.sort()
        label_vectors = np.array(label_vectors)
        image_features = StandardScaler().fit_transform(image_features)

        num_of_images = label_vectors.shape[0]
        combinations_list = np.array(list(combinations(range(num_of_images), 2)))
        similarity_matrix_lin_list = dict()

        for i, j in combinations_list:
            similarity_matrix_lin_list[
                get_key_for_indices(i, j)
            ] = distance_metric_multilabel(
                label_vectors[i], label_vectors[j], image_features[i], image_features[j]
            )
        # self.logger.info("calculating similarity {},{},{}".format(i, j, sim))

        for i in range(num_of_images):
            similarity_matrix_lin_list[get_key_for_indices(i, i)] = 1.0

        series = pd.Series(similarity_matrix_lin_list)
        matrix_df = pd.DataFrame(
            {"col_major_index": series.index, "similarity": series.values}
        )
        matrix_df.to_csv(similarity_matrix_file)
        self.logger.debug(
            "Done calculating the similarity matrix stored at: {}".format(
                similarity_matrix_file
            )
        )
        assert (
            self.n_features == image_features.shape[1]
        ), "Number of features not correct in dataset {}".format(
            image_features.shape[1]
        )
        return image_features, label_vectors



from os.path import dirname
from os.path import realpath

from setuptools import find_packages
from setuptools import setup

with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

DISTNAME = "csrank"
DESCRIPTION = "Context-sensitive ranking"
MAINTAINER = "Karlson Pfannschmidt"
MAINTAINER_EMAIL = "kiudee@mail.upb.de"

PROJECT_ROOT = dirname(realpath(__file__))

if __name__ == "__main__":
    setup(
        name=DISTNAME,
        version="1.2.0",
        maintainer=MAINTAINER,
        maintainer_email=MAINTAINER_EMAIL,
        classifiers=[
            "Development Status :: 4 - Beta",
            "Intended Audience :: Developers",
            "Intended Audience :: Science/Research",
            "License :: OSI Approved :: Apache Software License",
            "Natural Language :: English",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Topic :: Scientific/Engineering",
            "Topic :: Software Development",
        ],
        description=DESCRIPTION,
        packages=find_packages(),
        install_requires=[
            "numpy>=1.12.1",
            "scipy>=0.19.0",
            "scikit-learn>=0.18.2",
            "scikit-optimize>=0.4",
            "pandas>=0.22",
            "h5py>=2.7",
            "docopt>=0.6.0",
            "joblib>=0.9.4",
            "tqdm>=4.11.2",
            "keras>=2.3",
            # Pick either CPU or GPU version of tensorflow:
            "tensorflow>=1.5,<2.0",
            # tensorflow-gpu>=1.0.1"
        ],
        extras_require={
            "data": [
                "psycopg2-binary>=2.7",  # database access
                "pandas>=0.22",
                "h5py>=2.7",
                "pygmo>=2.7",
            ],
            "probabilistic": ["pymc3>=3.8", "theano>=1.0"],
        },
        package_data={"notebooks": ["*"]},
        include_package_data=True,
        long_description=readme + "\n\n" + history,
        url="https://github.com/kiudee/cs-ranking",
    )



from datetime import datetime
from datetime import timedelta
import inspect
import logging
import os
from pathlib import Path
import re
import sys

from csrank.metrics import make_ndcg_at_k_loss
from csrank.metrics import zero_one_accuracy
from csrank.metrics import zero_one_rank_loss
from csrank.metrics_np import f1_measure
from csrank.metrics_np import hamming
from csrank.metrics_np import instance_informedness
from csrank.metrics_np import precision
from csrank.metrics_np import recall
from csrank.metrics_np import subset_01_loss
from csrank.metrics_np import zero_one_accuracy_np

__all__ = [
    "create_dir_recursively",
    "duration_till_now",
    "print_dictionary",
    "rename_file_if_exist",
    "seconds_to_time",
    "time_from_now",
    "get_duration_seconds",
    "setup_logging",
    "progress_bar",
]


metrics_on_predictions = [
    f1_measure,
    precision,
    recall,
    subset_01_loss,
    hamming,
    instance_informedness,
    zero_one_rank_loss,
    zero_one_accuracy,
    make_ndcg_at_k_loss,
    zero_one_accuracy_np,
]


class MissingExtraError(ImportError):
    """Indicates an ImportError that can be fixed by installing an extra"""

    def __init__(self, package, extra):
        super(MissingExtraError, self).__init__(
            f"Could not import the optional dependency {package}. "
            f'Please install it or specify the "{extra}" extra when installing this package.'
        )


def progress_bar(count, total, status=""):
    bar_len = 60
    filled_len = int(round(bar_len * count / float(total)))

    bar = "=" * filled_len + "-" * (bar_len - filled_len)

    sys.stdout.write("[%s] %s/%s ...%s\r" % (bar, count, total, status))
    sys.stdout.flush()


def print_dictionary(dictionary):
    output = "\n"
    for key, value in dictionary.items():
        output = output + str(key) + " => " + str(value) + "\n"
    return output


def create_dir_recursively(path, is_file_path=False):
    if is_file_path:
        path = os.path.dirname(path)
    if not os.path.exists(path):
        os.makedirs(path, exist_ok=True)


def rename_file_if_exist(file_path):
    my_file = Path(file_path)
    try:
        extension = "." + file_path.split(".")[1]
    except IndexError:
        extension = ""
    path = file_path.split(".")[0]
    i = 1
    while my_file.is_file():
        file_path = path + str(i) + extension
        my_file = Path(file_path)
        i += 1
    return file_path


def get_duration_seconds(duration):
    time = int(re.findall(r"\d+", duration)[0])
    d = duration.split(str(time))[1].upper()
    options = {"D": 24 * 60 * 60, "H": 60 * 60, "M": 60}
    return options[d] * time


def duration_till_now(start):
    return (datetime.now() - start).total_seconds()


def time_from_now(target_time_sec):
    base_datetime = datetime.now()
    delta = timedelta(seconds=target_time_sec)
    target_date = base_datetime + delta
    return target_date.strftime("%Y-%m-%d %H:%M:%S")


def seconds_to_time(target_time_sec):
    return str(timedelta(seconds=target_time_sec))


def convert_to_loss(loss_function):
    def loss(y_true, y_pred):
        return 1.0 - loss_function(y_true, y_pred)

    return loss


def setup_logging(log_path=None, level=logging.DEBUG):
    """Function setup as many logging for the experiments"""
    if log_path is None:
        dirname = os.path.dirname(
            os.path.abspath(inspect.getfile(inspect.currentframe()))
        )
        dirname = os.path.dirname(dirname)
        log_path = os.path.join(dirname, "experiments", "logs", "logs.log")
        create_dir_recursively(log_path, True)
    logging.basicConfig(
        filename=log_path,
        level=level,
        format="%(asctime)s %(name)s %(levelname)-8s %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )
    logger = logging.getLogger("SetupLogger")
    logger.info("log file path: {}".format(log_path))



import inspect
import os

import numpy as np
import sklearn.preprocessing as sklearn_preprocessing

try:
    import pandas as pd
except ImportError:
    from csrank.util import MissingExtraError

    raise MissingExtraError("pandas", "data")

DATASET_FOLDER = "university_dataset"
UNIVERSITY_ = "university_"
WORLD_RANK = "world_rank"
COUNTRY = "country"
UNIVERSITY_NAME = "university_name"
years = [2012, 2013, 2014, 2015]
MIN_UNIV_RANK_LENGTH = 95

__all__ = ["univerisity_dataset_for_year", "MIN_UNIV_RANK_LENGTH", "years"]


def num(s):
    try:
        return float(s)
    except ValueError:
        try:
            return int(s)
        except ValueError:
            return s


def cal_mean(w):
    try:
        if "-" in w:
            return (float(str(w).split("-")[0]) + float(str(w).split("-")[1])) / 2
        else:
            return num(str(w))
    except TypeError:
        return num(str(w))


def create_joined_dataset():
    dir_name = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
    datasetCwur = pd.read_csv(
        os.path.join(dir_name, DATASET_FOLDER, "cwurData.csv")
    ).rename(columns={"institution": UNIVERSITY_NAME})
    datasetTimes = pd.read_csv(os.path.join(dir_name, DATASET_FOLDER, "timesData.csv"))
    datasetShanghai = pd.read_csv(
        os.path.join(dir_name, DATASET_FOLDER, "shanghaiData.csv")
    )
    columns = np.concatenate(
        (
            datasetCwur.columns[3:-1],
            datasetTimes.columns[3:-1],
            datasetShanghai.columns[3:-1],
        )
    ).flatten()
    columns = np.insert(columns, 0, UNIVERSITY_NAME)
    columns = np.insert(columns, 0, COUNTRY)
    columns = np.insert(columns, len(columns), WORLD_RANK)
    # columns = np.insert(columns, len(columns), WORLD_RANK+'_array')
    for year in years:
        df1 = datasetCwur.loc[datasetCwur["year"] == year]
        df2 = datasetTimes.loc[datasetTimes["year"] == year]
        df3 = datasetShanghai.loc[datasetShanghai["year"] == year]
        list1 = list(df1[UNIVERSITY_NAME])
        list2 = list(df2[UNIVERSITY_NAME])
        list3 = list(df3[UNIVERSITY_NAME])
        names = list(set().union(list1, list2, list3))
        dataYear = []

        for name in names:
            row1 = df1.loc[df1[UNIVERSITY_NAME] == name]
            row2 = df2.loc[df2[UNIVERSITY_NAME] == name]
            row3 = df3.loc[df3[UNIVERSITY_NAME] == name]
            a = len(row1) != 0
            b = len(row2) != 0
            c = len(row3) != 0
            conditions = np.array([a, b, c])

            if np.sum(conditions) >= 2:
                country = "none"
                world_ranks = []
                if b:
                    country = row2.iloc[0][2]
                if a:
                    country = row1.iloc[0][2]
                oneRow = np.array([name, country])
                if a:
                    row = [num(s) for s in row1.iloc[0][3:-1]]
                    oneRow = np.append(oneRow, row)
                    world_ranks.append(row1.iloc[0][0])
                else:
                    zeros = np.zeros(10)
                    zeros[:] = np.inf
                    oneRow = np.append(oneRow, zeros)

                if b:
                    row = [num(s) for s in row2.iloc[0][3:-1]]
                    row[-2] = float(str(row[-2]).split("%")[0])
                    try:
                        row[-1] = float(str(row[-1]).split(" : ")[0]) / float(
                            str(row[-1]).split(" : ")[1]
                        )
                    except IndexError:
                        row[-1] = np.inf
                    except ZeroDivisionError:
                        row[-1] = 1.0
                    oneRow = np.append(oneRow, row)
                    world_ranks.append(row2.iloc[0][0])
                else:
                    zeros = np.zeros(10)
                    zeros[:] = np.inf
                    oneRow = np.append(oneRow, zeros)

                if c:
                    row = [num(s) for s in row3.iloc[0][3:-1]]
                    oneRow = np.append(oneRow, row)
                    world_ranks.append(row3.iloc[0][0])

                else:
                    zeros = np.zeros(7)
                    zeros[:] = np.inf
                    oneRow = np.append(oneRow, zeros)
                world_rank = np.mean(np.array([cal_mean(w) for w in world_ranks]))
                oneRow = np.append(oneRow, [world_rank])
                # oneRow = np.append(oneRow, str(world_ranks))
                dataYear.append(oneRow)
        dataframeYear = pd.DataFrame(dataYear, columns=columns)
        dataframeYear = dataframeYear.convert_objects(convert_numeric=True)
        total_score = np.array(dataframeYear[["total_score", "score"]])[:, 0:2]
        total_score[np.isnan(total_score)] = 0.0
        total_score[np.isinf(total_score)] = 0.0
        dataframeYear["total_score"] = np.mean(total_score, axis=1)
        dataframeYear.drop("score", axis=1, inplace=True)
        dataframeYear.drop("broad_impact", axis=1, inplace=True)
        dataframeYear = dataframeYear.sort_values(by=[WORLD_RANK], ascending=[True])
        dataframeYear.to_csv(
            os.path.join(dir_name, DATASET_FOLDER, UNIVERSITY_ + str(year) + ".csv")
        )


# Get features and rankings for given hotel_dataset for a city
def univerisity_dataset_for_year(year):
    dir_name = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))

    if int(year) not in years:
        raise ValueError(
            f"Data is not present for the year, university dataset is present only for: {set(years)}"
        )
    path = os.path.join(dir_name, DATASET_FOLDER, UNIVERSITY_ + str(year) + ".csv")
    dataframe = pd.read_csv(path)
    headers = np.array(dataframe.columns)
    headers[1], headers[2] = headers[2], headers[1]
    dataframe = dataframe[headers]
    size = len(headers)
    for cat_cols in dataframe.select_dtypes(["object"]).columns:
        dataframe[cat_cols] = dataframe[cat_cols].astype("category")
        dataframe[cat_cols] = pd.Categorical.from_array(dataframe[cat_cols]).codes
        dataframe[cat_cols] = dataframe[cat_cols].replace(-1, np.NaN)

    array = dataframe.as_matrix()
    features = array[:, 2 : size - 1]
    headers = headers[2 : size - 1]
    for i in range(len(headers)):
        col = features[:, i]
        non_nans = [f for f in col if not (np.isinf(f) or np.isnan(f))]
        mean = np.nanmean(non_nans)
        col[np.isnan(col)] = mean
        col[np.isinf(col)] = mean

    # if (np.isnan(np.sum(features))):
    #     imp = sklearn_preprocessing.Imputer()
    #     features = imp.fit_transform(features).T
    actual_ranks = array[:, size - 1]
    scaler = sklearn_preprocessing.MinMaxScaler()
    normalized_ranks = np.array(actual_ranks)
    normalized_ranks = scaler.fit_transform(normalized_ranks.reshape(-1, 1)).T[0]
    # get features and filtered features
    features = np.array(
        [np.log(np.array(features[:, i]) + 1) for i in range(len(features[0]))]
    )
    scaler = sklearn_preprocessing.StandardScaler()
    features = scaler.fit_transform(features).T
    actual_ranks = np.arange(len(actual_ranks))

    features = features[None, :]
    actual_ranks = actual_ranks[None, :]
    normalized_ranks = normalized_ranks[None, :]

    return headers, features, actual_ranks, normalized_ranks



import logging

from sklearn.utils import check_random_state

from csrank.constants import CHOICE_FUNCTION
from .util import sub_sampling_choices_from_relevance
from ..letor_ranking_dataset_reader import LetorRankingDatasetReader


class LetorRankingChoiceDatasetReader(LetorRankingDatasetReader):
    def __init__(self, random_state=None, n_objects=5, **kwargs):
        super(LetorRankingChoiceDatasetReader, self).__init__(
            learning_problem=CHOICE_FUNCTION, **kwargs
        )
        self.logger = logging.getLogger(LetorRankingChoiceDatasetReader.__name__)
        self.random_state = check_random_state(random_state)
        self.n_objects = n_objects
        self.__load_dataset__()

    def sub_sampling_function(self, X, Y):
        return sub_sampling_choices_from_relevance(Xt=X, Yt=Y, n_objects=self.n_objects)

    def splitter(self, iter):
        pass

    def get_train_test_datasets(self, n_datasets):
        pass



from abc import ABCMeta

from keras import Input
from keras import Model
from keras.regularizers import l2
import numpy as np

from csrank import FATENetworkCore
from csrank import FATEObjectRanker
from csrank.tests.test_ranking import optimizer


def test_construction_core():
    n_objects = 3
    n_features = 2

    # Create mock class:

    class MockClass(FATENetworkCore, metaclass=ABCMeta):
        def set_tunable_parameters(self, **point):
            super().set_tunable_parameters(**point)

        def predict_scores(self, X, **kwargs):
            pass

        def _predict_scores_fixed(self, X, **kwargs):
            pass

        def predict(self, *args, **kwargs):
            pass

        def fit(self, *args, **kwargs):
            pass

    grc = MockClass()
    grc._construct_layers(
        activation=grc.activation,
        kernel_initializer=grc.kernel_initializer,
        kernel_regularizer=grc.kernel_regularizer,
    )
    input_layer = Input(shape=(n_objects, n_features))
    scores = grc.join_input_layers(input_layer, None, n_layers=0, n_objects=n_objects)

    model = Model(inputs=input_layer, outputs=scores)
    model.compile(loss="mse", optimizer=grc.optimizer)
    X = np.random.randn(100, n_objects, n_features)
    y = X.sum(axis=2)
    model.fit(x=X, y=y, verbose=0)
    params = {
        "n_hidden_joint_units": 2,
        "n_hidden_joint_layers": 10,
        "reg_strength": 1e-3,
        "learning_rate": 1e-1,
        "batch_size": 32,
    }
    grc.set_tunable_parameters(**params)
    assert grc.n_hidden_joint_units == params["n_hidden_joint_units"]
    assert grc.n_hidden_joint_layers == params["n_hidden_joint_layers"]
    assert grc.batch_size == params["batch_size"]
    rtol = 1e-2
    atol = 1e-4
    key = "learning_rate" if "learning_rate" in grc.optimizer.get_config() else "lr"
    learning_rate = grc.optimizer.get_config().get(key, 0.0)
    assert np.isclose(
        learning_rate, params["learning_rate"], rtol=rtol, atol=atol, equal_nan=False
    )
    config = grc.kernel_regularizer.get_config()
    val1 = np.isclose(
        config["l1"], params["reg_strength"], rtol=rtol, atol=atol, equal_nan=False
    )
    val2 = np.isclose(
        config["l2"], params["reg_strength"], rtol=rtol, atol=atol, equal_nan=False
    )
    assert val1 or val2


def test_fate_object_ranker_fixed_generator():
    def trivial_ranking_problem_generator():
        while True:
            rand = np.random.RandomState(123)
            x = rand.randn(10, 5, 1)
            y_true = x.argsort(axis=1).argsort(axis=1).squeeze(axis=-1)
            yield x, y_true

    fate = FATEObjectRanker(
        n_hidden_joint_layers=1,
        n_hidden_set_layers=1,
        n_hidden_joint_units=5,
        n_hidden_set_units=5,
        kernel_regularizer=l2(1e-4),
        optimizer=optimizer,
    )
    fate.fit_generator(
        generator=trivial_ranking_problem_generator(),
        epochs=1,
        validation_split=0,
        verbose=False,
        steps_per_epoch=10,
    )



import os

from keras.metrics import binary_accuracy
import numpy as np
import pytest

from csrank.tuning import check_learner_class
from ..tuning import ParameterOptimizer

OPTIMIZER_PATH = os.path.join(os.getcwd(), "opt")

n_instances = 500
n_features = 3


@pytest.fixture
def optimizer():
    from ..tunable import Tunable

    class RankerStub(Tunable):
        def fit(self, X, Y, **kwargs):
            self.seed = int(np.sum(list(self.__dict__.values())))

        def predict(self, X, **kwargs):
            random_state = np.random.RandomState(self.seed)
            weight = random_state.rand(n_features, 2)
            scores = np.dot(X, weight) / np.dot(X, weight).sum(axis=1)[:, None]
            return scores.argmax(axis=1)

        def set_tunable_parameters(self, **point):
            self.__dict__.update(point)

        def __call__(self, X, *args, **kwargs):
            return self.predict(X, **kwargs)

    ranker = RankerStub()

    rankers = [RankerStub() for _ in range(2)]
    test_params = {
        rankers[0]: dict(a=(1.0, 4.0)),
        ranker: dict(b=(4.0, 7.0), c=(7.0, 10.0)),
        rankers[1]: dict(d=(10.0, 13.0)),
    }

    opt = ParameterOptimizer(
        learner=ranker,
        optimizer_path=OPTIMIZER_PATH,
        tunable_parameter_ranges=test_params,
        ranker_params=dict(),
        validation_loss=binary_accuracy,
    )
    return opt, rankers, test_params


@pytest.fixture(scope="module")
def trivial_ranking_problem():
    random_state = np.random.RandomState(123)
    x = random_state.rand(n_instances, n_features)
    y_true = random_state.randint(0, 2, n_instances)
    return x, y_true


def check(value, bound):
    if bound[0] <= value <= bound[1]:
        return True
    return False


def test_parameter_optimizer(trivial_ranking_problem, optimizer):
    x, y = trivial_ranking_problem
    opt, rankers, test_params = optimizer
    opt.fit(x, y, n_iter=10)
    os.remove(OPTIMIZER_PATH)
    for point in opt.opt.Xi:
        i = 0
        for param_ranges in test_params.values():
            for key, bound in param_ranges.items():
                assert check(point[i], bound)
                i += 1


def test_set_parameters(optimizer):
    opt, rankers, test_params = optimizer
    point = list(range(4))
    opt._set_new_parameters(point)
    i = 0
    for ranker in test_params.keys():
        try:
            check_learner_class(ranker)
        except AttributeError as exc:
            pytest.fail(exc, pytrace=True)
        for param in test_params[ranker]:
            assert ranker.__dict__[param] == point[i]
            i += 1



from itertools import combinations
import logging

import numpy as np
from sklearn.utils import check_random_state

from csrank.learner import Learner
import csrank.numpy_util as npu
import csrank.theano_util as ttu
from csrank.util import print_dictionary
from .discrete_choice import DiscreteObjectChooser
from .likelihoods import create_weight_dictionary
from .likelihoods import fit_pymc3_model
from .likelihoods import likelihood_dict
from .likelihoods import LogLikelihood

try:
    import pymc3 as pm
    from pymc3.variational.callbacks import CheckParametersConvergence
except ImportError:
    from csrank.util import MissingExtraError

    raise MissingExtraError("pymc3", "probabilistic")

try:
    import theano
    from theano import tensor as tt
except ImportError:
    from csrank.util import MissingExtraError

    raise MissingExtraError("theano", "probabilistic")


class PairedCombinatorialLogit(DiscreteObjectChooser, Learner):
    def __init__(
        self,
        loss_function="",
        regularization="l2",
        alpha=5e-2,
        random_state=None,
        **kwd,
    ):
        """
            Create an instance of the Paired Combinatorial Logit model for learning the discrete choice function. This
            model considering each pair of objects as a different nest allowing unique covariances for each pair of objects,
            and each object is a member of :math:`n - 1` nests. This model structure is 1-layer of hierarchy and the
            :math:`\\lambda` for each nest :math:`B_k` signifies the degree of independence and  :math:`1-\\lambda` signifies
            the correlations between the object in it. We learn two weight vectors and the  :math:`\\lambda s`.
                * **weights** (:math:`w`): Weights to get the utility of the object :math:`Y_i = U(x_i) = w \\cdot x_i`
                * **lambda_k** (:math:`\\lambda_k`): Lambda for nest nest :math:`B_k` for correlations between the obejcts.

            The probability of choosing an object :math:`x_i` from the given query set :math:`Q` is defined by product
            of choosing the nest in which :math:`x_i` exists and then choosing the the object from the nest.

            .. math::

                P(x_i \\lvert Q) = P_i = \\sum_{\\substack{B_k \\in \\mathcal{B} \\ i \\in B_k}}P_{i \\lvert B_k} P_{B_k} \\enspace ,


            The discrete choice for the given query set :math:`Q` is defined as:

            .. math::

                dc(Q) := \\operatorname{argmax}_{x_i \\in Q }  \\; P(x_i \\lvert Q)

            Parameters
            ----------
            n_objects: int
                Number of objects in each query set
            n_nests : int range : [2,n_objects/2]
                The number of nests/subsets in which the objects are divided
            loss_function : string , {‘categorical_crossentropy’, ‘binary_crossentropy’, ’categorical_hinge’}
                Loss function to be used for the discrete choice decision from the query set
            regularization : string, {‘l1’, ‘l2’}, string
               Regularizer function (L1 or L2) applied to the `kernel` weights matrix
            alpha: float (range : [0,1])
                The lower bound of the correlations between the objects in a nest
            random_state : int or object
                Numpy random state
            **kwargs
                Keyword arguments for the algorithms

            References
            ----------
                [1] Kenneth E Train. „Discrete choice methods with simulation“. In: Cambridge university press, 2009. Chap GEV, pp. 87–111.

                [2] Kenneth Train. Qualitative choice analysis. Cambridge, MA: MIT Press, 1986

                [3] Chaushie Chu. „A paired combinatorial logit model for travel demand analysis“. In: Proceedings of the fifth world conference on transportation research. Vol. 4.1989, pp. 295–309
        """
        self.logger = logging.getLogger(PairedCombinatorialLogit.__name__)
        self.alpha = alpha
        self.random_state = random_state
        self.loss_function = likelihood_dict.get(loss_function, None)
        known_regularization_functions = {"l1", "l2"}
        if regularization not in known_regularization_functions:
            raise ValueError(
                f"Regularization function {regularization} is unknown. Must be one of {known_regularization_functions}"
            )
        self.regularization = regularization
        self._config = None
        self.model = None
        self.trace = None
        self.trace_vi = None
        self.Xt = None
        self.Yt = None
        self.p = None

    @property
    def model_configuration(self):
        """
            Constructs the dictionary containing the priors for the weight vectors for the model according to the
            regularization function. The parameters are:
                * **weights** : Weights to evaluates the utility of the objects

            For ``l1`` regularization the priors are:

            .. math::

                \\text{mu}_w \\sim \\text{Normal}(\\text{mu}=0, \\text{sd}=5.0) \\\\
                \\text{b}_w \\sim \\text{HalfCauchy}(\\beta=1.0) \\\\
                \\text{weights} \\sim \\text{Laplace}(\\text{mu}=\\text{mu}_w, \\text{b}=\\text{b}_w)

            For ``l2`` regularization the priors are:

            .. math::

                \\text{mu}_w \\sim \\text{Normal}(\\text{mu}=0, \\text{sd}=5.0) \\\\
                \\text{sd}_w \\sim \\text{HalfCauchy}(\\beta=1.0) \\\\
                \\text{weights} \\sim \\text{Normal}(\\text{mu}=\\text{mu}_w, \\text{sd}=\\text{sd}_w)

            Returns
            -------
                configuration : dict
                    Dictionary containing the priors applies on the weights
        """
        if self._config is None:
            if self.regularization == "l2":
                weight = pm.Normal
                prior = "sd"
            elif self.regularization == "l1":
                weight = pm.Laplace
                prior = "b"
            self._config = {
                "weights": [
                    weight,
                    {
                        "mu": (pm.Normal, {"mu": 0, "sd": 5}),
                        prior: (pm.HalfCauchy, {"beta": 1}),
                    },
                ]
            }
            self.logger.info(
                "Creating model with config {}".format(print_dictionary(self._config))
            )
        return self._config

    #

    def get_probabilities(self, utility, lambda_k):
        """
            This method calculates the probability of choosing an object from the query set using the following parameters of the model which are used:

                * **weights** (:math:`w`): Weights to get the utility of the object :math:`Y_i = U(x_i) = w \\cdot x_i`
                * **lambda_k** (:math:`\\lambda_k`): Lambda is the measure of independence amongst the obejcts in the nest :math:`B_k`

            The probability of choosing the object  :math:`x_i` from the query set :math:`Q`:

            .. math::
                    P_i = \\sum_{j \\in I \\setminus i} P_{{i} \\lvert {ij}} P_{ij} \\enspace where, \\\\
                    P_{i \\lvert ij} = \\frac{\\boldsymbol{e}^{^{Y_i} /_{\\lambda_{ij}}}}{\\boldsymbol{e}^{^{Y_i} /_{\\lambda_{ij}}} + \\boldsymbol{e}^{^{Y_j} /_{\\lambda_{ij}}}} \\enspace ,\\\\
                    P_{ij} = \\frac{{\\left( \\boldsymbol{e}^{^{V_i}/{\\lambda_{ij}}} + \\boldsymbol{e}^{^{V_j}/{\\lambda_{ij}}}  \\right)}^{\\lambda_{ij}}}{\\sum_{k=1}^{n-1} \\sum_{\\ell = k + 1}^{n} {\\left( \\boldsymbol{e}^{^{V_k}/{\\lambda_{k\\ell}}} + \\boldsymbol{e}^{^{V_{\\ell}}/{\\lambda_{k\\ell}}}  \\right)}^{\\lambda_{k\\ell}}}


            Parameters
            ----------
            utility : theano tensor
                (n_instances, n_objects)
                Utility :math:`Y_i` of the objects :math:`x_i \\in Q` in the query sets
            lambda_k : theano tensor (range : [alpha, 1.0])
                (n_nests)
                Measure of independence amongst the obejcts in each nests

            Returns
            -------
            p : theano tensor
                (n_instances, n_objects)
                Choice probabilities :math:`P_i` of the objects :math:`x_i \\in Q` in the query sets

        """
        n_objects = self.n_objects_fit_
        nests_indices = self.nests_indices
        n_nests = self.n_nests
        lambdas = tt.ones((n_objects, n_objects), dtype=np.float)
        for i, p in enumerate(nests_indices):
            r = [p[0], p[1]]
            c = [p[1], p[0]]
            lambdas = tt.set_subtensor(lambdas[r, c], lambda_k[i])
        uti_per_nest = tt.transpose(utility[:, None, :] / lambdas, (0, 2, 1))
        ind = np.array([[[i1, i2], [i2, i1]] for i1, i2 in nests_indices])
        ind = ind.reshape(2 * n_nests, 2)
        x = uti_per_nest[:, ind[:, 0], ind[:, 1]].reshape((-1, 2))
        log_sum_exp_nest = ttu.logsumexp(x).reshape((-1, n_nests))
        pnk = tt.exp(
            log_sum_exp_nest * lambda_k - ttu.logsumexp(log_sum_exp_nest * lambda_k)
        )
        p = tt.zeros(tuple(utility.shape), dtype=float)
        for i in range(n_nests):
            i1, i2 = nests_indices[i]
            x1 = tt.exp(uti_per_nest[:, i1, i2] - log_sum_exp_nest[:, i]) * pnk[:, i]
            x2 = np.exp(uti_per_nest[:, i2, i1] - log_sum_exp_nest[:, i]) * pnk[:, i]
            p = tt.set_subtensor(p[:, i1], p[:, i1] + x1)
            p = tt.set_subtensor(p[:, i2], p[:, i2] + x2)
        return p

    def _get_probabilities_np(self, utility, lambda_k):
        n_objects = self.n_objects_fit_
        nests_indices = self.nests_indices
        n_nests = self.n_nests
        temp_lambdas = np.ones((n_objects, n_objects), lambda_k.dtype)
        temp_lambdas[nests_indices[:, 0], nests_indices[:, 1]] = temp_lambdas.T[
            nests_indices[:, 0], nests_indices[:, 1]
        ] = lambda_k
        uti_per_nest = np.transpose((utility[:, None] / temp_lambdas), (0, 2, 1))
        ind = np.array([[[i1, i2], [i2, i1]] for i1, i2 in nests_indices])
        ind = ind.reshape(2 * n_nests, 2)
        x = uti_per_nest[:, ind[:, 0], ind[:, 1]].reshape(-1, 2)
        log_sum_exp_nest = npu.logsumexp(x).reshape(-1, n_nests)
        pnk = np.exp(
            log_sum_exp_nest * lambda_k - npu.logsumexp(log_sum_exp_nest * lambda_k)
        )
        p = np.zeros(tuple(utility.shape), dtype=float)
        for i in range(n_nests):
            i1, i2 = nests_indices[i]
            p[:, i1] += (
                np.exp(uti_per_nest[:, i1, i2] - log_sum_exp_nest[:, i]) * pnk[:, i]
            )
            p[:, i2] += (
                np.exp(uti_per_nest[:, i2, i1] - log_sum_exp_nest[:, i]) * pnk[:, i]
            )
        return p

    def construct_model(self, X, Y):
        """
            Constructs the nested logit model by applying priors on weight vectors **weights** as per :meth:`model_configuration`.
            Then we apply a uniform prior to the :math:`\\lambda s`, i.e. :math:`\\lambda s \\sim Uniform(\\text{alpha}, 1.0)`.
            The probability of choosing the object :math:`x_i` from the query set :math:`Q = \\{x_1, \\ldots ,x_n\\}` is
            evaluated in :meth:`get_probabilities`.

            Parameters
            ----------
            X : numpy array
                (n_instances, n_objects, n_features)
                Feature vectors of the objects
            Y : numpy array
                (n_instances, n_objects)
                Preferences in the form of discrete choices for given objects

            Returns
            -------
             model : pymc3 Model :class:`pm.Model`
        """
        with pm.Model() as self.model:
            self.Xt = theano.shared(X)
            self.Yt = theano.shared(Y)
            shapes = {"weights": self.n_object_features_fit_}
            weights_dict = create_weight_dictionary(self.model_configuration, shapes)
            lambda_k = pm.Uniform("lambda_k", self.alpha, 1.0, shape=self.n_nests)
            utility = tt.dot(self.Xt, weights_dict["weights"])
            self.p = self.get_probabilities(utility, lambda_k)
            LogLikelihood(
                "yl", loss_func=self.loss_function, p=self.p, observed=self.Yt
            )
        self.logger.info("Model construction completed")

    def fit(
        self,
        X,
        Y,
        sampler="variational",
        tune=500,
        draws=500,
        vi_params={
            "n": 20000,
            "method": "advi",
            "callbacks": [CheckParametersConvergence()],
        },
        **kwargs,
    ):
        """
           Fit a paired combinatorial logit  model on the provided set of queries X and choices Y of those objects. The
           provided queries and corresponding preferences are of a fixed size (numpy arrays). For learning this network
           the categorical cross entropy loss function for each object :math:`x_i \\in Q` is defined as:

           .. math::

               C_{i} =  -y(i)\\log(P_i) \\enspace,

           where :math:`y` is ground-truth discrete choice vector of the objects in the given query set :math:`Q`.
           The value :math:`y(i) = 1` if object :math:`x_i` is chosen else :math:`y(i) = 0`.

           Parameters
           ----------
           X : numpy array (n_instances, n_objects, n_features)
               Feature vectors of the objects
           Y : numpy array (n_instances, n_objects)
               Choices for given objects in the query
           sampler : {‘variational’, ‘metropolis’, ‘nuts’}, string
               The sampler used to estimate the posterior mean and mass matrix from the trace

                   * **variational** : Run inference methods to estimate posterior mean and diagonal mass matrix
                   * **metropolis** : Use the MAP as starting point and Metropolis-Hastings sampler
                   * **nuts** : Use the No-U-Turn sampler
           vi_params : dict
               The parameters for the **variational** inference method
           draws : int
               The number of samples to draw. Defaults to 500. The number of tuned samples are discarded by default
           tune : int
               Number of iterations to tune, defaults to 500. Ignored when using 'SMC'. Samplers adjust
               the step sizes, scalings or similar during tuning. Tuning samples will be drawn in addition
               to the number specified in the `draws` argument, and will be discarded unless
               `discard_tuned_samples` is set to False.
           **kwargs :
               Keyword arguments for the fit function of :meth:`pymc3.fit`or :meth:`pymc3.sample`
       """
        self.random_state_ = check_random_state(self.random_state)
        _n_instances, self.n_objects_fit_, self.n_object_features_fit_ = X.shape
        self.nests_indices = np.array(
            list(combinations(np.arange(self.n_objects_fit_), 2))
        )
        self.n_nests = len(self.nests_indices)
        self.construct_model(X, Y)
        fit_pymc3_model(self, sampler, draws, tune, vi_params, **kwargs)

    def _predict_scores_fixed(self, X, **kwargs):
        mean_trace = dict(pm.summary(self.trace)["mean"])
        weights = np.array(
            [
                mean_trace["weights[{}]".format(i)]
                for i in range(self.n_object_features_fit_)
            ]
        )
        lambda_k = np.array(
            [mean_trace["lambda_k[{}]".format(i)] for i in range(self.n_nests)]
        )
        utility = np.dot(X, weights)
        p = self._get_probabilities_np(utility, lambda_k)
        return p

    def predict(self, X, **kwargs):
        return super().predict(X, **kwargs)

    def predict_scores(self, X, **kwargs):
        return super().predict_scores(X, **kwargs)

    def predict_for_scores(self, scores, **kwargs):
        return DiscreteObjectChooser.predict_for_scores(self, scores, **kwargs)

    def set_tunable_parameters(
        self, alpha=5e-2, loss_function=None, regularization="l2", **point
    ):
        """
            Set tunable parameters of the Paired Combinatorial logit model to the values provided.

            Parameters
            ----------
            alpha: float (range : [0,1])
                The lower bound of the correlations between the objects in a nest
            loss_function : string , {‘categorical_crossentropy’, ‘binary_crossentropy’, ’categorical_hinge’}
                Loss function to be used for the discrete choice decision from the query set
            regularization : string, {‘l1’, ‘l2’}, string
               Regularizer function (L1 or L2) applied to the `kernel` weights matrix
            point: dict
                Dictionary containing parameter values which are not tuned for the network
        """
        if alpha is not None:
            self.alpha = alpha
        if loss_function is not None:
            if loss_function not in likelihood_dict:
                raise ValueError(
                    f"Loss function {loss_function} is unknown. Must be one of {set(likelihood_dict.keys())}"
                )
            self.loss_function = likelihood_dict.get(loss_function, None)
        self.regularization = regularization
        self.model = None
        self.trace = None
        self.trace_vi = None
        self.Xt = None
        self.Yt = None
        self.p = None
        self._config = None
        if len(point) > 0:
            self.logger.warning(
                "This ranking algorithm does not support tunable parameters"
                " called: {}".format(print_dictionary(point))
            )
